@@ -5,14 +5,15 @@ from typing import Callable, Any, Optional
 
 from mcdreforged import constant
 from mcdreforged.executor.thread_executor import ThreadExecutor
+from mcdreforged.utils import misc_util
 from mcdreforged.utils.logger import DebugOption
 
 
 class Priority:
-	# High priority
+	# High priority, executes first
 	REGULAR = 0
 	INFO = 20
-	# Low priority
+	# Low priority, executes last
 
 
 class TaskData:
@@ -72,16 +73,12 @@ class TaskExecutor(ThreadExecutor):
 		else:
 			self.task_queue.put_nowait(data)
 
-	def enqueue_regular_task(self, func: Callable[[], Any]):
+	def add_regular_task(self, func: Callable[[], Any], *, wait=False):
+		if wait:
+			func = misc_util.WaitableCallable(func)
 		self.task_queue.put(TaskData(func, Priority.REGULAR))
-
-	def execute_or_enqueue(self, func: Callable[[], Any], *, wait=False):
-		if self.is_on_thread():
-			func()
-		else:
-			self.enqueue_regular_task(func)
-			if wait:
-				self.wait_till_finish_all_task()
+		if wait:
+			func.wait()
 
 	def get_this_tick_time(self) -> float:
 		"""
