@@ -48,10 +48,10 @@ class PluginManager:
 		self.thread_pool = PluginThreadPool(self.mcdr_server, max_thread=constant.PLUGIN_THREAD_POOL_SIZE)
 
 		# thread local storage, to store current plugin
-		self.tls = ThreadLocalStorage()
+		self.__tls = ThreadLocalStorage()
 
 		# plugin manipulation lock
-		self.mani_lock = threading.RLock()
+		self.__mani_lock = threading.RLock()
 
 		file_util.touch_directory(PLUGIN_CONFIG_DIRECTORY)
 
@@ -64,9 +64,12 @@ class PluginManager:
 		Get current executing plugin in this thread
 		:param thread: If specified, it should be a Thread instance. Then it will return the executing plugin in the given thread
 		"""
-		stack = self.tls.get(self.TLS_PLUGIN_KEY, None, thread=thread)  # type: Deque[AbstractPlugin]
+		stack = self.__tls.get(self.TLS_PLUGIN_KEY, None, thread=thread)  # type: Deque[AbstractPlugin]
 		return stack[-1] if stack is not None else None
 
+	def get_plugin_amount(self) -> int:
+		return len(self.plugins)
+
 	def get_all_plugins(self) -> List[AbstractPlugin]:
 		return list(self.plugins.values())
 
@@ -97,15 +100,15 @@ class PluginManager:
 
 	@contextmanager
 	def with_plugin_context(self, plugin: AbstractPlugin):
-		stack = self.tls.get(self.TLS_PLUGIN_KEY, default=collections.deque())  # type: Deque[AbstractPlugin]
+		stack = self.__tls.get(self.TLS_PLUGIN_KEY, default=collections.deque())  # type: Deque[AbstractPlugin]
 		stack.append(plugin)
-		self.tls.put(self.TLS_PLUGIN_KEY, stack)
+		self.__tls.put(self.TLS_PLUGIN_KEY, stack)
 		try:
 			yield
 		finally:
 			stack.pop()
 			if len(stack) == 0:
-				self.tls.pop(self.TLS_PLUGIN_KEY)
+				self.__tls.pop(self.TLS_PLUGIN_KEY)
 
 	def contains_plugin_file(self, file_path: str) -> bool:
 		return file_path in self.plugin_file_path
@@ -126,7 +129,8 @@ class PluginManager:
 
 	def register_permanent_plugins(self):
 		self.__add_permanent_plugin(MCDReforgedPlugin(self))
-		self.__update_registry()  # not really necessary, but in case
+		self.__sort_plugins_by_id()
+		self.__update_registry()
 
 	# ------------------------------------------------
 	#   Actual operations that add / remove a plugin
@@ -246,14 +250,17 @@ class PluginManager:
 	def __collect_and_process_new_plugins(self, filter: Callable[[str], bool]) -> SingleOperationResult:
 		result = SingleOperationResult()
 		for plugin_directory in self.plugin_directories:
-			file_list = file_util.list_file_with_suffix(plugin_directory, constant.PLUGIN_FILE_SUFFIX)
-			for file_path in file_list:
-				if not self.contains_plugin_file(file_path) and filter(file_path):
-					plugin = self.__load_plugin(file_path)
-					if plugin is None:
-						result.fail(file_path)
-					else:
-						result.succeed(plugin)
+			if os.path.isdir(plugin_directory):
+				file_list = file_util.list_file_with_suffix(plugin_directory, constant.PLUGIN_FILE_SUFFIX)
+				for file_path in file_list:
+					if not self.contains_plugin_file(file_path) and filter(file_path):
+						plugin = self.__load_plugin(file_path)
+						if plugin is None:
+							result.fail(file_path)
+						else:
+							result.succeed(plugin)
+			else:
+				self.logger.warning('Plugin directory "{}" not found'.format(plugin_directory))
 		return result
 
 	def __collect_and_remove_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
@@ -331,6 +338,12 @@ class PluginManager:
 		# reload_result		READY				UNLOADING
 		# dep_chk_result	LOADED / READY		UNLOADING
 
+		# Stuffs execution order in self.mcdr_server.task_executor
+		# 1. PLUGIN_LOADED event listeners callbacks
+		# 2. self.__update_registry (watchdog disabled)
+		# 3. PLUGIN_LOADED, PLUGIN_REMOVED event listeners callbacks
+		# This method will wait until step 2 finishes before quiting
+
 		self.registry_storage.clear()  # in case plugin invokes dispatch_event during on_load. dont let them trigger listeners
 
 		for plugin in load_result.success_list + reload_result.success_list:
@@ -343,6 +356,8 @@ class PluginManager:
 				if isinstance(plugin, RegularPlugin):
 					plugin.receive_event(MCDRPluginEvents.PLUGIN_LOADED, (plugin.old_module_instance,))
 
+		self.mcdr_server.task_executor.add_regular_task(self.__update_registry)
+
 		for plugin in unload_result.success_list + unload_result.failed_list + reload_result.failed_list + dependency_check_result.failed_list:
 			plugin.assert_state({PluginState.UNLOADING})
 			# plugins might just be newly loaded but failed on dependency check, dont dispatch event to them
@@ -356,17 +371,17 @@ class PluginManager:
 			plugin.assert_state({PluginState.READY})
 
 		self.__sort_plugins_by_id()
-		self.__update_registry()
 
 	def __sort_plugins_by_id(self):
 		self.plugins = dict(sorted(self.plugins.items(), key=lambda item: item[0]))
 
 	def __update_registry(self):
-		self.registry_storage.clear()
-		for plugin in self.get_all_plugins():
-			self.registry_storage.collect(plugin.plugin_registry)
-		self.registry_storage.arrange()
-		self.mcdr_server.on_plugin_changed()
+		with self.mcdr_server.watch_dog.pausing():
+			self.registry_storage.clear()
+			for plugin in self.get_all_plugins():
+				self.registry_storage.collect(plugin.plugin_registry)
+			self.registry_storage.arrange()
+			self.mcdr_server.on_plugin_registry_changed()
 
 	def __refresh_plugins(self, reload_filter: Callable[[RegularPlugin], bool]):
 		unload_result = self.__collect_and_remove_plugins(lambda plugin: not plugin.file_exists())
@@ -379,25 +394,25 @@ class PluginManager:
 	# --------------
 
 	def load_plugin(self, file_path: str):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.entered', file_path))
 			load_result = self.__collect_and_process_new_plugins(lambda fp: fp == file_path)
 			self.__post_plugin_process(load_result=load_result)
 
 	def unload_plugin(self, plugin: RegularPlugin):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.unload_plugin.entered', plugin))
 			unload_result = self.__collect_and_remove_plugins(lambda plg: True, specific=plugin)
 			self.__post_plugin_process(unload_result=unload_result)
 
 	def reload_plugin(self, plugin: RegularPlugin):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.entered', plugin))
 			reload_result = self.__reload_ready_plugins(lambda plg: True, specific=plugin)
 			self.__post_plugin_process(reload_result=reload_result)
 
 	def enable_plugin(self, file_path: str):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.enable_plugin.entered', file_path))
 			new_file_path = string_util.remove_suffix(file_path, constant.DISABLED_PLUGIN_FILE_SUFFIX)
 			if os.path.isfile(file_path):
@@ -405,19 +420,19 @@ class PluginManager:
 				self.load_plugin(new_file_path)
 
 	def disable_plugin(self, plugin: RegularPlugin):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.disable_plugin.entered', plugin))
 			self.unload_plugin(plugin)
 			if os.path.isfile(plugin.file_path):
 				os.rename(plugin.file_path, plugin.file_path + constant.DISABLED_PLUGIN_FILE_SUFFIX)
 
 	def refresh_all_plugins(self):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_all_plugins.entered'))
 			return self.__refresh_plugins(lambda plg: True)
 
 	def refresh_changed_plugins(self):
-		with self.mani_lock:
+		with self.__mani_lock:
 			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_changed_plugins.entered'))
 			return self.__refresh_plugins(lambda plg: plg.file_changed())
 
@@ -426,24 +441,40 @@ class PluginManager:
 	# ----------------
 
 	def __dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...]):
+		"""
+		Event dispatch logic implementation
+		"""
 		self.logger.debug('Dispatching {} with args ({})'.format(event, ', '.join([type(arg).__name__ for arg in args])), option=DebugOption.PLUGIN)
 		for listener in self.registry_storage.event_listeners.get(event.id, []):
-			self.trigger_listener(listener, args)
+			self.__trigger_listener(listener, args)
 
-	def dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...], *, on_executor_thread=True, wait=False):
-		if on_executor_thread:
-			self.mcdr_server.task_executor.execute_or_enqueue(lambda: self.__dispatch_event(event, args), wait=wait)
-		else:  # on thread
-			self.__dispatch_event(event, args)
-
-	def trigger_listener(self, listener: EventListener, args: Tuple[Any, ...]):
+	def __trigger_listener(self, listener: EventListener, args: Tuple[Any, ...]):
 		"""
-		The terminated entry for triggering a listener
+		Event listener triggering implementation
 		The server_interface parameter will be automatically added as the 1st parameter
 		"""
-		# self.thread_pool.add_task(lambda: listener.execute(*args), listener.plugin)
 		with self.with_plugin_context(listener.plugin):
 			try:
 				listener.execute(self.mcdr_server.server_interface, *args)
 			except:
 				self.logger.exception('Error invoking listener {}'.format(listener))
+
+	def dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...], *, run_async=True, wait=False):
+		"""
+		Event dispatching interface. Event will be added into the task queue of the task executor thread
+		"""
+		if run_async:
+			self.mcdr_server.task_executor.add_regular_task(lambda: self.__dispatch_event(event, args), wait=wait)
+		else:
+			self.__dispatch_event(event, args)
+
+	def trigger_listener(self, listener: EventListener, args: Tuple[Any, ...], *, run_async=True, wait=False):
+		"""
+		Listener triggering interface. Event will be added into the task queue of the task executor thread
+		Should only be used in specific plugin listener triggering
+		"""
+		# self.thread_pool.add_task(lambda: listener.execute(*args), listener.plugin)
+		if run_async:
+			self.mcdr_server.task_executor.add_regular_task(lambda: self.__trigger_listener(listener, args), wait=wait)
+		else:
+			self.__trigger_listener(listener, args)
